@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2002 Zope Foundation and Contributors.
+# Copyright (c) 2002, 2018 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -18,7 +18,6 @@ import doctest
 import logging
 import os
 import sys
-import tempfile
 import unittest
 
 import ZConfig
@@ -30,7 +29,8 @@ from ZConfig.components.logger import loghandler
 from ZConfig._compat import NStringIO as StringIO
 from ZConfig._compat import maxsize
 
-from ZConfig.tests.support import TestHelper
+import ZConfig.tests.support
+import ZConfig.components.logger.tests.support
 
 
 class CustomFormatter(logging.Formatter):
@@ -50,69 +50,8 @@ def read_file(filename):
         return f.read()
 
 
-class LoggingTestHelper(TestHelper):
-
-    # Not derived from unittest.TestCase; some test runners seem to
-    # think that means this class contains tests.
-
-    # XXX This tries to save and restore the state of logging around
-    # the test.  Somewhat surgical; there may be a better way.
-
-    def setUp(self):
-        self._created = []
-        self._old_logger = logging.getLogger()
-        self._old_level = self._old_logger.level
-        self._old_handlers = self._old_logger.handlers[:]
-        self._old_logger.handlers[:] = []
-        self._old_logger.setLevel(logging.WARN)
-
-        self._old_logger_dict = logging.root.manager.loggerDict.copy()
-        logging.root.manager.loggerDict.clear()
-
-    def tearDown(self):
-        logging.root.manager.loggerDict.clear()
-        logging.root.manager.loggerDict.update(self._old_logger_dict)
-
-        for h in self._old_logger.handlers:
-            self._old_logger.removeHandler(h)
-        for h in self._old_handlers:
-            self._old_logger.addHandler(h) # pragma: no cover
-        self._old_logger.setLevel(self._old_level)
-
-        while self._created:
-            os.unlink(self._created.pop())
-
-        self.assertEqual(loghandler._reopenable_handlers, [])
-        loghandler.closeFiles()
-        loghandler._reopenable_handlers == []
-
-    def mktemp(self):
-        fd, fn = tempfile.mkstemp()
-        os.close(fd)
-        self._created.append(fn)
-        return fn
-
-    def move(self, fn):
-        nfn = self.mktemp()
-        os.rename(fn, nfn)
-        return nfn
-
-    _schema = None
-
-    def get_schema(self):
-        if self._schema is None:
-            sio = StringIO(self._schematext)
-            self.__class__._schema = ZConfig.loadSchemaFile(sio)
-        return self._schema
-
-    def get_config(self, text):
-        conf, handler = ZConfig.loadConfigFile(self.get_schema(),
-                                               StringIO(text))
-        self.assertTrue(not handler)
-        return conf
-
-
-class TestConfig(LoggingTestHelper, unittest.TestCase):
+class TestConfig(ZConfig.components.logger.tests.support.LoggingTestHelper,
+                 unittest.TestCase):
 
     _schematext = """
       <schema>
@@ -170,6 +109,45 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
         logfile = logger.handlers[0]
         self.assertEqual(logfile.level, logging.DEBUG)
         self.assertTrue(isinstance(logfile, loghandler.FileHandler))
+        self.assertFalse(logfile.delay)
+        self.assertIsNotNone(logfile.stream)
+        logger.removeHandler(logfile)
+        logfile.close()
+
+    def test_with_encoded(self):
+        fn = self.mktemp()
+        logger = self.check_simple_logger("<eventlog>\n"
+                                          "  <logfile>\n"
+                                          "    path %s\n"
+                                          "    level debug\n"
+                                          "    encoding shift-jis\n"
+                                          "  </logfile>\n"
+                                          "</eventlog>" % fn)
+        logfile = logger.handlers[0]
+        self.assertEqual(logfile.level, logging.DEBUG)
+        self.assertTrue(isinstance(logfile, loghandler.FileHandler))
+        self.assertFalse(logfile.delay)
+        self.assertIsNotNone(logfile.stream)
+        self.assertEqual(logfile.stream.encoding, "shift-jis")
+        logger.removeHandler(logfile)
+        logfile.close()
+
+    def test_with_logfile_delayed(self):
+        fn = self.mktemp()
+        logger = self.check_simple_logger("<eventlog>\n"
+                                          "  <logfile>\n"
+                                          "    path %s\n"
+                                          "    level debug\n"
+                                          "    delay true\n"
+                                          "  </logfile>\n"
+                                          "</eventlog>" % fn)
+        logfile = logger.handlers[0]
+        self.assertEqual(logfile.level, logging.DEBUG)
+        self.assertTrue(isinstance(logfile, loghandler.FileHandler))
+        self.assertTrue(logfile.delay)
+        self.assertIsNone(logfile.stream)
+        logger.info("this is a test")
+        self.assertIsNotNone(logfile.stream)
         logger.removeHandler(logfile)
         logfile.close()
 
@@ -179,6 +157,18 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
     def test_with_stdout(self):
         self.check_standard_stream("stdout")
 
+    def test_delayed_stderr(self):
+        self.check_standard_stream_cannot_delay("stderr")
+
+    def test_delayed_stdout(self):
+        self.check_standard_stream_cannot_delay("stdout")
+
+    def test_encoded_stderr(self):
+        self.check_standard_stream_cannot_encode("stderr")
+
+    def test_encoded_stdout(self):
+        self.check_standard_stream_cannot_encode("stdout")
+
     def test_with_rotating_logfile(self):
         fn = self.mktemp()
         logger = self.check_simple_logger("<eventlog>\n"
@@ -212,15 +202,15 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
         self.assertEqual(logfile.level, logging.DEBUG)
         self.assertEqual(logfile.backupCount, 11)
         self.assertEqual(logfile.interval, 86400*3)
-        self.assertTrue(isinstance(logfile, loghandler.TimedRotatingFileHandler))
+        self.assertIsInstance(logfile, loghandler.TimedRotatingFileHandler)
         logger.removeHandler(logfile)
         logfile.close()
 
     def test_with_timed_rotating_logfile_and_size_should_fail(self):
         fn = self.mktemp()
         self.assertRaises(
-            ValueError,
-            self.check_simple_logger,
+            ZConfig.DataConversionError,
+            self.check_simple_logger_factory,
             "<eventlog>\n"
             "  <logfile>\n"
             "    path %s\n"
@@ -233,9 +223,9 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
 
         # Mising old-files
         self.assertRaisesRegex(
-            ValueError,
+            ZConfig.DataConversionError,
             "old-files must be set",
-            self.check_simple_logger,
+            self.check_simple_logger_factory,
             "<eventlog>\n"
             "  <logfile>\n"
             "    path %s\n"
@@ -246,9 +236,9 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
             "</eventlog>" % fn)
 
         self.assertRaisesRegex(
-            ValueError,
+            ZConfig.DataConversionError,
             "max-bytes or when must be set",
-            self.check_simple_logger,
+            self.check_simple_logger_factory,
             "<eventlog>\n"
             "  <logfile>\n"
             "    path %s\n"
@@ -258,13 +248,12 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
             "  </logfile>\n"
             "</eventlog>" % fn)
 
-
     def test_with_rotating_logfile_and_STD_should_fail(self):
         for path in ('STDERR', 'STDOUT'):
             for param in ('old-files 10', 'max-size 5mb'):
                 self.assertRaises(
-                    ValueError,
-                    self.check_simple_logger,
+                    ZConfig.DataConversionError,
+                    self.check_simple_logger_factory,
                     "<eventlog>\n"
                     "  <logfile>\n"
                     "    path %s\n"
@@ -274,7 +263,6 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
                     "  </logfile>\n"
                     "</eventlog>" % (path, param))
 
-
     def check_standard_stream(self, name):
         old_stream = getattr(sys, name)
         conf = self.get_config("""
@@ -296,22 +284,53 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
         finally:
             setattr(sys, name, old_stream)
         logger.warning("woohoo!")
+        self.assertIs(logger.handlers[0].stream, sio)
         self.assertTrue(sio.getvalue().find("woohoo!") >= 0)
 
+    def check_standard_stream_cannot_delay(self, name):
+        with self.assertRaises(ZConfig.DataConversionError) as cm:
+            self.get_config("""
+                <eventlog>
+                  <logfile>
+                    level info
+                    path %s
+                    delay true
+                  </logfile>
+                </eventlog>
+                """ % name.upper())
+
+        self.assertIn("cannot delay opening %s" % name.upper(),
+                      str(cm.exception))
+
+    def check_standard_stream_cannot_encode(self, name):
+        with self.assertRaises(ZConfig.DataConversionError) as cm:
+            self.get_config("""
+                <eventlog>
+                  <logfile>
+                    level info
+                    path %s
+                    encoding utf-8
+                  </logfile>
+                </eventlog>
+                """ % name.upper())
+        self.assertIn("cannot specify encoding for %s" % name.upper(),
+                      str(cm.exception))
+
     def test_custom_formatter(self):
+        clsname = __name__ + '.CustomFormatter'
         old_stream = sys.stdout
-        conf = self.get_config("""
-        <eventlog>
-        <logfile>
-        formatter ZConfig.components.logger.tests.test_logger.CustomFormatter
-        level info
-        path STDOUT
-        </logfile>
-        </eventlog>
-        """)
         sio = StringIO()
         sys.stdout = sio
         try:
+            conf = self.get_config("""
+                <eventlog>
+                  <logfile>
+                    formatter %s
+                    level info
+                    path STDOUT
+                  </logfile>
+                </eventlog>
+                """ % clsname)
             logger = conf.eventlog()
         finally:
             sys.stdout = old_stream
@@ -333,10 +352,10 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
         syslog = logger.handlers[0]
         self.assertEqual(syslog.level, logging.ERROR)
         self.assertTrue(isinstance(syslog, loghandler.SysLogHandler))
-        syslog.close() # avoid ResourceWarning
+        syslog.close()  # avoid ResourceWarning
         try:
-            syslog.socket.close() # ResourceWarning under 3.2
-        except socket.SocketError: # pragma: no cover
+            syslog.socket.close()  # ResourceWarning under 3.2
+        except socket.SocketError:  # pragma: no cover
             pass
 
     def test_with_http_logger_localhost(self):
@@ -410,26 +429,35 @@ class TestConfig(LoggingTestHelper, unittest.TestCase):
         self.assertEqual(handler.mailport, 487)
 
     def test_with_email_notifier_with_invalid_credentials(self):
-        self.assertRaises(ValueError,
-                          self.check_simple_logger,
-                          "<eventlog>\n"
-                          "  <email-notifier>\n"
-                          "    to sysadmin@example.com\n"
-                          "    from zlog-user@example.com\n"
-                          "    level fatal\n"
-                          "    smtp-username john\n"
-                          "  </email-notifier>\n"
-                          "</eventlog>")
-        self.assertRaises(ValueError,
-                          self.check_simple_logger,
-                          "<eventlog>\n"
-                          "  <email-notifier>\n"
-                          "    to sysadmin@example.com\n"
-                          "    from zlog-user@example.com\n"
-                          "    level fatal\n"
-                          "    smtp-password john\n"
-                          "  </email-notifier>\n"
-                          "</eventlog>")
+        # smtp-username without smtp-password
+        with self.assertRaises(ZConfig.DataConversionError) as cm:
+            self.check_simple_logger_factory(
+                "<eventlog>\n"
+                "  <email-notifier>\n"
+                "    to sysadmin@example.com\n"
+                "    from zlog-user@example.com\n"
+                "    level fatal\n"
+                "    smtp-username john\n"
+                "  </email-notifier>\n"
+                "</eventlog>")
+        self.assertIn(
+            'both smtp-username and smtp-password or none must be given',
+            str(cm.exception))
+
+        # smtp-password without smtp-username
+        with self.assertRaises(ZConfig.DataConversionError) as cm:
+            self.check_simple_logger_factory(
+                "<eventlog>\n"
+                "  <email-notifier>\n"
+                "    to sysadmin@example.com\n"
+                "    from zlog-user@example.com\n"
+                "    level fatal\n"
+                "    smtp-password john\n"
+                "  </email-notifier>\n"
+                "</eventlog>")
+        self.assertIn(
+            'both smtp-username and smtp-password or none must be given',
+            str(cm.exception))
 
     def check_simple_logger_factory(self, text, level=logging.INFO):
         conf = self.get_config(text)
@@ -458,7 +486,9 @@ else:
     _RotateTestBase = unittest.TestCase
 
 
-class TestReopeningRotatingLogfiles(LoggingTestHelper, _RotateTestBase):
+class TestReopeningRotatingLogfiles(
+        ZConfig.components.logger.tests.support.LoggingTestHelper,
+        _RotateTestBase):
 
     # These tests should not be run on Windows.
 
@@ -682,17 +712,86 @@ class TestReopeningLogfiles(TestReopeningRotatingLogfiles):
         h.release = lambda: calls.append("release")
 
         h.reopen()
-        h.close()
-
         self.assertEqual(calls, ["acquire", "release"])
+        del calls[:]
 
-class TestFunctions(TestHelper, unittest.TestCase):
+        # FileHandler.close() does acquire/release, and invokes
+        # StreamHandler.flush(), which does the same. Since the lock is
+        # recursive, that's ok.
+        #
+        h.close()
+        self.assertEqual(calls, ["acquire", "acquire", "release", "release"])
 
-    def test_log_format_bad(self):
-        self.assertRaisesRegex(ValueError,
-                               "Invalid log format string",
-                               handlers.log_format,
-                               "%{no-such-key}s")
+    def test_with_logfile_delayed_reopened(self):
+        fn = self.mktemp()
+        conf = self.get_config("<logger>\n"
+                               "  <logfile>\n"
+                               "    path %s\n"
+                               "    level debug\n"
+                               "    delay true\n"
+                               "    encoding shift-jis\n"
+                               "  </logfile>\n"
+                               "</logger>" % fn)
+        logger = conf.loggers[0]()
+        logfile = logger.handlers[0]
+        self.assertTrue(logfile.delay)
+        self.assertIsNone(logfile.stream)
+        logger.info("this is a test")
+        self.assertIsNotNone(logfile.stream)
+        self.assertEqual(logfile.stream.encoding, "shift-jis")
+
+        # After reopening, we expect the stream to be reset, to be
+        # opened at the next event to be logged:
+        logfile.reopen()
+        self.assertIsNone(logfile.stream)
+        logger.info("this is another test")
+        self.assertIsNotNone(logfile.stream)
+        self.assertEqual(logfile.stream.encoding, "shift-jis")
+        logger.removeHandler(logfile)
+        logfile.close()
+
+
+class TestFunctions(ZConfig.tests.support.TestHelper, unittest.TestCase):
+
+    def test_log_format_bad_syntax_1(self):
+        with self.assertRaises(ValueError) as cm:
+            # Really, disallowed character following '%':
+            handlers.log_format('%{no-such-key}s')
+        self.assertEqual(str(cm.exception),
+                         'Invalid log format string %{no-such-key}s')
+
+    def test_log_format_bad_syntax_2(self):
+        with self.assertRaises(ValueError) as cm:
+            # Missing trailing 's':
+            handlers.log_format('%(levelname)')
+        self.assertEqual(str(cm.exception),
+                         'Invalid log format string %(levelname)')
+
+    def test_log_format_unknown_key(self):
+        with self.assertRaises(ValueError) as cm:
+            handlers.log_format('%(no-such-key)s')
+        self.assertEqual(str(cm.exception),
+                         'Invalid log format string %(no-such-key)s')
+
+    def test_log_format_ok(self):
+        fmt = handlers.log_format(r'\n\t\b\f\r')
+        self.assertEqual(fmt, '\n\t\b\f\r')
+
+    def test_log_format_func_name(self):
+        fmt = '%(funcName)s'
+        self.assertEqual(handlers.log_format(fmt), fmt)
+
+    def test_log_format_levelno_integer(self):
+        fmt = '%(levelno)2d'
+        self.assertEqual(handlers.log_format(fmt), fmt)
+
+    def test_log_format_msecs_float(self):
+        fmt = '%(msecs)03.0f'
+        self.assertEqual(handlers.log_format(fmt), fmt)
+
+    def test_log_format_relative_created_float(self):
+        fmt = '%(relativeCreated)+.3f'
+        self.assertEqual(handlers.log_format(fmt), fmt)
 
     def test_resolve_deep(self):
         old_mod = None
@@ -727,9 +826,12 @@ class TestFunctions(TestHelper, unittest.TestCase):
     def test_close_files(self):
         class F(object):
             closed = 0
+
             def close(self):
                 self.closed += 1
+
         f = F()
+
         def wr():
             return f
 
@@ -758,6 +860,9 @@ class TestFunctions(TestHelper, unittest.TestCase):
             self.assertEqual(convert(name), convert(name.upper()))
         self.assertRaises(ValueError, convert, "hopefully-not-a-valid-value")
         self.assertEqual(convert('10'), 10)
+        self.assertRaises(ValueError, convert, '-1000')
+        self.assertRaises(ValueError, convert, '-1')
+        self.assertRaises(ValueError, convert, '51')
         self.assertRaises(ValueError, convert, '100')
 
     def test_http_method(self):
@@ -788,8 +893,10 @@ class TestStartupHandler(unittest.TestCase):
         self.assertEqual(maxsize, handler.capacity)
 
         records = []
+
         def handle(record):
             records.append(record)
+
         handle.handle = handle
 
         handler.flushBufferTo(handle)
@@ -801,6 +908,7 @@ class TestStartupHandler(unittest.TestCase):
 
         del handle.handle
 
+
 def test_logger_convenience_function_and_ommiting_name_to_get_root_logger():
     """
 
@@ -853,11 +961,13 @@ We'll configure the rot logger and a non-root logger.
 
     """
 
+
 def test_suite():
     return unittest.TestSuite([
         unittest.defaultTestLoader.loadTestsFromName(__name__),
         doctest.DocTestSuite()
     ])
 
+
 if __name__ == '__main__':
     unittest.main(defaultTest="test_suite")
